#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>

// for_each_process definition
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
#include <linux/sched/signal.h>
#include <linux/sched/task.h>
#else
#include <linux/sched.h>
#endif

#ifndef RHEL_RELEASE_CODE
#define RHEL_RELEASE_CODE          0
#define RHEL_RELEASE_VERSION(a, b) (((a) << 8) + (b))
#endif

#ifndef module_init
#define module_init(initfunc) int init_module(void) { return initfunc(); }
#endif
#ifndef module_exit
#define module_exit(exitfunc) void cleanup_module(void) { exitfunc(); }
#endif
#ifndef __init
#define __init
#endif
#ifndef __exit
#define __exit
#endif
#ifndef MODULE_LICENSE
#define MODULE_LICENSE(licstr)
#endif
#ifndef MODULE_AUTHOR
#define MODULE_AUTHOR(authstr)
#endif
#ifndef MODULE_DESCRIPTION
#define MODULE_DESCRIPTION(descstr)
#endif

#ifndef _LINUX_
#define _LINUX_
#endif
#include "AmiFltKern.h"
#define LOGPREF "Amicon: "

#define LIN_DEVNAME         "/tmp/AmiVpn"
#define LIN_EVENTNAME       "AmiVpnEvent"
#define LIN_RULESFNAME      "/usr/lib/Amicon_ip-client/rules.dat"
#define LIN_FSGUARDCFGFNAME "/usr/lib/Amicon_ip-client/fsguard.cfg"
#define LIN_FSGUARDLOGFNAME "/usr/lib/Amicon_ip-client/fsguard.log"

#if defined(CONFIG_X86_32) || defined(CONFIG_X86_64)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 2, 0)
#include <asm/fpu/api.h>
#else
#include <asm/i387.h>
#endif
void KERN_fpu_begin(void) { kernel_fpu_begin(); }
void KERN_fpu_end(void)   { kernel_fpu_end();   }
#endif
#if defined(CONFIG_ARM64)
#include <asm/neon.h>
void KERN_fpu_begin(void) { kernel_neon_begin(); }
void KERN_fpu_end(void)   { kernel_neon_end();   }
#endif

///// -> Strings **********************************************
void __calling_convention __attribute__ ((format (printf, 1, 2))) KERN_printk(const char *fmt, ...) {
  va_list args; va_start(args, fmt); vprintk(fmt, args); va_end(args);
}
void __calling_convention __attribute__ ((format (printf, 3, 4))) KERN_snprintf(char *str, int str_len, const char *fmt, ...) {
  va_list args; va_start(args, fmt); vsnprintf(str, str_len, fmt, args); va_end(args);
}
void __calling_convention KERN_strncpy(char *dst, const char *src, int dst_len) { strncpy(dst, src, dst_len - 1); dst[dst_len - 1] = 0; }
void __calling_convention KERN_strncat(char *dst, const char *src, int dst_len) { strncat(dst, src, dst_len - strlen(dst) - 1); dst[dst_len - 1] = 0; }
int  __calling_convention KERN_strcmp(const char *dst, const char *src) { return strcmp(dst, src); }
int  __calling_convention KERN_strlen(const char *str) { return strlen(str); }
void __calling_convention KERN_memset(void *mem, char val, int size) { memset(mem, val, size); }
void __calling_convention KERN_memcpy(void *dst, const void *src, int size) { memcpy(dst, src, size); }
void __calling_convention KERN_memmove(void *dst, const void *src, int size) { memmove(dst, src, size); }
///// <- Strings **********************************************

///// -> Module initialization ********************************
static int  __init KERN_init_module(void)    { return Ami_init_module(); }
static void __exit KERN_cleanup_module(void) { Ami_cleanup_module(); flush_scheduled_work(); }
///// <- Module initialization ********************************

///// -> Network hooks ****************************************
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ip.h>
#include <net/ip.h>
#include <linux/if_arp.h>
#include <asm/uaccess.h>


#define AMICON_FILTER_TAG          ('A'*16777216 + 'F'*65536 + 'L'*256 + 'T') // 'AFLT'
#define AMICON_FILTER_CB_TAG(skb)  AMICON_FILTER_TAG
#define AMICON_FILTER_SKB_TAG(skb) (*(int *)&((skb)->cb[sizeof((skb)->cb) - sizeof(int)]))

static int KERN_netbuf_setour(KERN_netbuf netbuf) {
  AMICON_FILTER_SKB_TAG(((struct sk_buff *)netbuf)) = AMICON_FILTER_CB_TAG(((struct sk_buff *)netbuf));
  return 0;
}
static int KERN_netbuf_checkour(KERN_netbuf netbuf) {
  if(AMICON_FILTER_SKB_TAG(((struct sk_buff *)netbuf)) == AMICON_FILTER_CB_TAG(((struct sk_buff *)netbuf))) return 0;
  return -1;
}

#ifndef NIC_ETHERNET // See CliFpsu.h
enum
{
  NIC_ETHERNET = 0, // 14b header
  NIC_NOMACS,       // no  macs, only 2-bytes proto
  NIC_IP,           // no  header
  NIC_PPP,          // 2b  header
  NIC_MAX
};
#endif
#ifndef NON_BLOCK_RET // See CliFpsu.h
#define NON_BLOCK_RET        0 // Send,       OK answer
#define BLOCK_SUCCESFULL_RET 1 // Don't send, OK answer
#define INTERNAL_ERROR_RET   2 // Don't send, ERROR answer
#endif

static unsigned int KERN_nf_inputoutput(
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 0)
                                        void *priv,
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0)) || (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 0))
                                        const struct nf_hook_ops *ops,
#else
                                        unsigned int hooknum,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
struct sk_buff *skb,
#else
struct sk_buff **pskb,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 1, 0)
  const struct nf_hook_state *state)
#elif RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 2)
  const struct net_device *indev, const struct net_device *outdev, const struct nf_hook_state *state)
#else
  const struct net_device *indev, const struct net_device *outdev, int (*okfn)(struct sk_buff *))
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 0)
  unsigned int hooknum = state->hook;
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0)) || (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 0))
  unsigned int hooknum = ops->hooknum;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24)
  struct sk_buff *skb = *pskb;
#endif
  int ForSend =
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 25)
    !(hooknum == NF_INET_PRE_ROUTING);
#else
    !(hooknum == NF_IP_PRE_ROUTING);
#endif
  int Status;
  if(!KERN_netbuf_checkour(skb)) return NF_ACCEPT; // It's our packet
  Status = Ami_net_inputoutput(ForSend, (void *)skb, (void *)skb->dev);
  switch(Status) {
  case NON_BLOCK_RET:        Status = NF_ACCEPT; break;
  case BLOCK_SUCCESFULL_RET:
    if(ForSend) {            Status = NF_STOLEN; KERN_netbuf_free(skb); }
    else                     Status = NF_DROP;
    break;
  case INTERNAL_ERROR_RET:   Status = NF_DROP;   break;
  default:                   Status = NF_DROP;   break;
  };
  return Status;
}
static struct nf_hook_ops KERN_nf_incoming = { .hook = &KERN_nf_inputoutput,
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 4, 0)
.owner = THIS_MODULE,
#endif
.pf = PF_INET,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 25)
.hooknum  = NF_INET_PRE_ROUTING,
#else
.hooknum  = NF_IP_PRE_ROUTING,
#endif
.priority = NF_IP_PRI_FIRST
};
static struct nf_hook_ops KERN_nf_outcoming = { .hook = &KERN_nf_inputoutput,
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 4, 0)
.owner = THIS_MODULE,
#endif
.pf = PF_INET,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 25)
.hooknum  = NF_INET_POST_ROUTING,
#else
.hooknum  = NF_IP_POST_ROUTING,
#endif
.priority = NF_IP_PRI_FIRST
};
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0)
// Used source from kernel 4.12 /linux/net/netfilter/core.c
static int nf_register_hook(struct nf_hook_ops *reg) {
  struct net *net, *last;
  int         ret;
  rtnl_lock();
  for_each_net(net) { ret = nf_register_net_hook(net, reg); if(ret && ret != -ENOENT) {
    last = net;
    for_each_net(net) { if(net == last) break; nf_unregister_net_hook(net, reg); }
    rtnl_unlock();
    return ret;
  }}
  rtnl_unlock();
  return 0;
}
static void nf_unregister_hook(struct nf_hook_ops *reg) {
  struct net *net;
  rtnl_lock(); for_each_net(net) nf_unregister_net_hook(net, reg); rtnl_unlock();
}
#endif
void __calling_convention KERN_net_register_hooks(void) { nf_register_hook(&KERN_nf_incoming); nf_register_hook(&KERN_nf_outcoming); }
void __calling_convention KERN_net_unregister_hooks(void) { nf_unregister_hook(&KERN_nf_incoming); nf_unregister_hook(&KERN_nf_outcoming); }
///// <- Network hooks ****************************************

///// -> Memory ***********************************************
void * __calling_convention KERN_kmalloc(int size) { return kmalloc(size, GFP_ATOMIC); }
void __calling_convention KERN_kfree(void *pmem) { if(pmem) kfree(pmem); }
///// <- Memory ***********************************************

///// -> Network device ***************************************
int __calling_convention KERN_netdev_getaddinfolen(void) { return 0; /* No needed in Linux */ }
///// <- Network device ***************************************

///// -> Network buffer ***************************************
KERN_netbuf __calling_convention KERN_netbuf_alloc(KERN_netdevice netdev, void *buffer, int buffer_len, int ForSend) {
  // Allocate
  struct sk_buff *skb;
  struct net_device *dev = (struct net_device *)netdev;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
  skb = dev_alloc_skb(buffer_len); if(skb && ForSend) skb->dev = dev;
#else
  if(ForSend) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 32)
#ifndef NET_IP_ALIGN
#define NET_IP_ALIGN 2
#endif
    skb = netdev_alloc_skb(dev, buffer_len + LL_RESERVED_SPACE(dev));
    if(NET_IP_ALIGN && skb) skb_reserve(skb, NET_IP_ALIGN);
#else
    skb = netdev_alloc_skb_ip_align(dev, buffer_len + LL_RESERVED_SPACE(dev));
#endif
    if(LL_RESERVED_SPACE(dev) && skb)
      skb_reserve(skb, LL_RESERVED_SPACE(dev));
  }
  else skb = dev_alloc_skb(buffer_len);
#endif
  if(!skb) return NULL;
  // Set data
  {
    unsigned char *pMac = NULL, *pIP = NULL;
    pMac = skb_put(skb, buffer_len); memcpy(pMac, buffer, buffer_len); pIP = pMac + ETH_HLEN;
    skb->ip_summed = CHECKSUM_UNNECESSARY;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 22)
    skb_set_mac_header(skb, 0);
    skb_set_network_header(skb, ETH_HLEN);
    if(ForSend) { if(!((struct iphdr *)pIP)->frag_off) skb_set_transport_header(skb, ETH_HLEN + ((struct iphdr *)pIP)->ihl * 4); }
#else
    skb->mac.raw = pMac; skb->nh.raw = pIP;
    if(ForSend) { if(!((struct iphdr *)pIP)->frag_off) skb->h.raw = pIP + ((struct iphdr *)pIP)->ihl * 4; }
#endif
    if(ForSend) { skb->pkt_type = PACKET_OUTGOING;
    skb->protocol = *((unsigned short *)((char *)buffer + ETH_HLEN - 2)); }
    else { skb->pkt_type = PACKET_HOST;
    skb->protocol = eth_type_trans(skb, dev); }
  }
  // Set timestamp
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0)
  {
    struct timespec64 timestamp;
    ktime_get_real_ts64(&timestamp);
    skb->tstamp = timespec64_to_ktime(timestamp);
  }
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
  {
    struct timeval timestamp;
    do_gettimeofday(&timestamp);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 22)
    skb->tstamp = timeval_to_ktime(timestamp);
#else
    skb_set_timestamp(skb, &timestamp);
#endif
  }
#else
  do_gettimeofday(&skb->stamp);
#endif
  return (KERN_netbuf)skb;
}
void __calling_convention KERN_netbuf_free(KERN_netbuf netbuf) { kfree_skb((struct sk_buff *)netbuf); }
int __calling_convention KERN_netbuf_len(KERN_netbuf netbuf) { return ((struct sk_buff *)netbuf)->len + ETH_HLEN; }
int __calling_convention KERN_netbuf_getdata(KERN_netbuf netbuf, char *buffer, int buffer_len) {
  if(buffer_len < ETH_HLEN) return 1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 22)
  if(skb_mac_header_was_set((struct sk_buff *)netbuf)) memcpy(buffer, skb_mac_header((struct sk_buff *)netbuf), ETH_HLEN);
#else
  if(((struct sk_buff *)netbuf)->mac.raw) memcpy(buffer, ((struct sk_buff *)netbuf)->mac.raw, ETH_HLEN);
#endif
  else { memset(buffer, 0, ETH_HLEN); *((unsigned short *)(buffer + ETH_HLEN - 2)) = htons(ETH_P_IP); }
  {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 22)
    int offset = skb_network_offset((struct sk_buff *)netbuf);
#else
    int offset = (char *)((struct sk_buff *)netbuf)->nh.raw - (char *)((struct sk_buff *)netbuf)->data;
#endif
    if(skb_copy_bits((struct sk_buff *)netbuf, offset, buffer + ETH_HLEN, buffer_len - ETH_HLEN)) return 1;
  }
  return 0;
}
struct work_struct_send_recv_data {
  struct work_struct work;
  KERN_netbuf netbuf;
  KERN_netdevice netdev;
  int ForSend;
};
static void worker_task_handler_fn(struct work_struct *work) {
  struct work_struct_send_recv_data *data = (struct work_struct_send_recv_data *)work;
  KERN_netbuf_setour(data->netbuf);
  if(data->ForSend) dev_queue_xmit((struct sk_buff *)data->netbuf);
  else {
    local_bh_disable(); // Disable softirq here to avoid race between timers and netif_receive_skb
    netif_receive_skb((struct sk_buff *)data->netbuf);
    local_bh_enable();
  }
  Ami_net_inputoutput_complete(data->ForSend, data->netbuf, data->netdev); // Complete now!
  // !!! netbuf was already freed in KERN_nf_inputoutput if it was needed!!!
  // !!! KERN_netbuf_free(data->netbuf);
  KERN_kfree(data);
}
static int __calling_convention KERN_netbuf_queue_recv_send(KERN_netbuf *netbuf, int netbufcount, KERN_netdevice netdev, int ForSend) {
  int netbufn; //, rc = 0;
  struct work_struct_send_recv_data *worker_task_data;
  for(netbufn = 0; netbufn < netbufcount; netbufn++) {
    worker_task_data = (struct work_struct_send_recv_data *)KERN_kmalloc(sizeof(*worker_task_data));
    if(worker_task_data) {
      INIT_WORK(&worker_task_data->work, worker_task_handler_fn);
      worker_task_data->netbuf = netbuf[netbufn];
      worker_task_data->netdev = netdev;
      worker_task_data->ForSend = ForSend;
      schedule_work(&worker_task_data->work);
    }
    else {
      KERN_netbuf_setour(netbuf[netbufn]);
      //rc =
      if(ForSend) dev_queue_xmit((struct sk_buff *)netbuf[netbufn]);
      else {
        local_bh_disable(); // Disable softirq here to avoid race between timers and netif_receive_skb
        netif_receive_skb((struct sk_buff *)netbuf[netbufn]);
        local_bh_enable();
      }
      Ami_net_inputoutput_complete(ForSend, netbuf[netbufn], netdev); // Complete now!
      // !!! netbuf was already freed in KERN_nf_inputoutput if it was needed!!!
      // !!! KERN_netbuf_free(netbuf[netbufn]);
    }
  }
  return 0; //rc;
}
int __calling_convention KERN_netbuf_queue_send(KERN_netbuf *netbuf, int netbufcount, KERN_netdevice netdev) {
  return KERN_netbuf_queue_recv_send(netbuf, netbufcount, netdev, 1);
}
int __calling_convention KERN_netbuf_queue_recv(KERN_netbuf *netbuf, int netbufcount, KERN_netdevice netdev) {
  return KERN_netbuf_queue_recv_send(netbuf, netbufcount, netdev, 0);
}
///// <- Network buffer ***************************************

///// -> Spin locks *******************************************
#include <linux/spinlock.h>
#ifdef __USE_IRQSPINLOCK__
void __calling_convention KERN_spin_lock_init(KERN_spin_lock *pLock) { if(!pLock) return; *pLock = (KERN_spin_lock)kmalloc(sizeof(spinlock_t) + sizeof(unsigned long), GFP_ATOMIC);
if(*pLock) { memset(*pLock, 0, sizeof(spinlock_t) + sizeof(unsigned long)); spin_lock_init((spinlock_t *)*pLock); }
}
void __calling_convention KERN_spin_lock_lock(KERN_spin_lock *pLock) { if(pLock) if(*pLock) spin_lock_irqsave((spinlock_t *)*pLock, *(unsigned long *)(*pLock + sizeof(spinlock_t))); }
void __calling_convention KERN_spin_lock_unlock(KERN_spin_lock *pLock) { if(pLock) if(*pLock) spin_unlock_irqrestore((spinlock_t *)*pLock, *(unsigned long *)(*pLock + sizeof(spinlock_t))); }
#else
void __calling_convention KERN_spin_lock_init(KERN_spin_lock *pLock) { if(!pLock) return; *pLock = (KERN_spin_lock)kmalloc(sizeof(spinlock_t), GFP_ATOMIC);
if(*pLock) { memset(*pLock, 0, sizeof(spinlock_t)); spin_lock_init((spinlock_t *)*pLock); }
}
void __calling_convention KERN_spin_lock_lock(KERN_spin_lock *pLock) { if(pLock) if(*pLock) spin_lock_bh((spinlock_t *)*pLock); }
void __calling_convention KERN_spin_lock_unlock(KERN_spin_lock *pLock) { if(pLock)if(*pLock) spin_unlock_bh((spinlock_t *)*pLock); }
#endif
void __calling_convention KERN_spin_lock_free(KERN_spin_lock *pLock) { if(pLock) if(*pLock) { kfree(*pLock); *pLock = NULL; } }
///// <- Spin locks *******************************************

///// -> Timer ************************************************
static struct timer_list netTimer;
static void timer_task_handler_fn(struct work_struct *work) {
  (void)work;

  Ami_DEV_NetTimerCallback();
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0)
static unsigned long netTimerVal = 0;
static void KERN_NetTimerCallback(struct timer_list *netTimerList) {
#else
static void KERN_NetTimerCallback(unsigned long netTimerVal) {
#endif
  static struct work_struct timer_work;
  static int timer_work_inited = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0)
  (void)netTimerList;
#endif
  if(!timer_work_inited) { INIT_WORK(&timer_work, timer_task_handler_fn); timer_work_inited++; }
  schedule_work(&timer_work); // Use work because of interrupt
  if(netTimerVal) mod_timer(&netTimer, jiffies + ((netTimerVal*HZ) / 1000)); }
void __calling_convention KERN_InitTimer(unsigned long timerval) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0)
  netTimerVal   = timerval; timer_setup(&netTimer, KERN_NetTimerCallback, 0);
#else
  netTimer.data = timerval; netTimer.function = KERN_NetTimerCallback; init_timer(&netTimer);
#endif
  netTimer.expires = jiffies + ((timerval*HZ) / 1000); add_timer(&netTimer);
}
void __calling_convention KERN_DestroyTimer(void) { del_timer(&netTimer); }
#include <linux/delay.h>
void __calling_convention KERN_Sleep(int ms) { msleep(ms); }
#include <linux/rtc.h>
void __calling_convention KERN_DrvGetCurFtime(int *sec, int *min, int *hour, int *mday, int *mon, int *year, unsigned long long *microsec) {
  struct rtc_time ctm;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
  time64_t ctime = ktime_get_real_seconds();
  rtc_time64_to_tm(ctime, &ctm);
#else
  time_t ctime = get_seconds();
  rtc_time_to_tm(ctime, &ctm);
#endif
  *sec = ctm.tm_sec; *min = ctm.tm_min; *hour = ctm.tm_hour; *mday = ctm.tm_mday; *mon = ctm.tm_mon; *year = ctm.tm_year + 1900;
  if(microsec) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 18, 0)
    *microsec = (unsigned long long)(ktime_get_real_ns() >> 10);
#else
    struct timespec mtm; getnstimeofday(&mtm);
    *microsec = ((unsigned long long)mtm.tv_sec) * 1000000 + (((unsigned long long)mtm.tv_nsec) >> 10);
#endif
  }
}
///// <- Timer ************************************************

///// -> Device ***********************************************
#include <net/genetlink.h>
#define VERSION_NR 1
enum { DRV_DEVICE_CMD_UNSPEC, DRV_DEVICE_CMD, DRV_DEVICE_CMD_MAX };
static int KERN_DeviceNetLinkSocketIoControl(struct sk_buff *skb, struct genl_info *info);
static struct nla_policy DRV_DEVICE_genl_policy[DRV_DEVICE_CMD_MAX] = { {.type = NLA_BINARY
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)
,.validation_type = NLA_VALIDATE_NONE
#endif
}, {.type = NLA_BINARY
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)
,.validation_type = NLA_VALIDATE_NONE
#endif
}, };
static struct genl_ops DRV_DEVICE_gnl_ops[] = { { .cmd = DRV_DEVICE_CMD, .flags = 0,
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 2, 0) && (RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(8, 0))
.policy = DRV_DEVICE_genl_policy,
#endif
.doit = KERN_DeviceNetLinkSocketIoControl, .dumpit = NULL, }, };
static struct genl_family DRV_DEVICE_gnl_family = { .hdrsize = 0,
.name = "", .version = VERSION_NR, .maxattr = DRV_DEVICE_CMD_MAX - 1,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 12))
.module = THIS_MODULE,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 2, 0)
.policy = DRV_DEVICE_genl_policy,
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)) || (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 2))
.ops = DRV_DEVICE_gnl_ops, .n_ops = ARRAY_SIZE(DRV_DEVICE_gnl_ops),
#else
.id = GENL_ID_GENERATE,
#endif
};
static char         *KERN_DeviceMessage = 0;
static unsigned int  KERN_DeviceMessage_len = 0;
static int KERN_DeviceNetLinkSocketIoControl(struct sk_buff *skb, struct genl_info *info) {
  struct nlattr *na;
  char          *mydata;
  unsigned int   ioctl_size = 0;
  unsigned int   ioctl_num = 0;
  void          *msg_head;
  if(!KERN_DeviceMessage || !KERN_DeviceMessage_len) { printk(LOGPREF "KERN_DeviceNetLinkSocketIoControl: not inited\n"); return 0; }
  if(!info) { printk(LOGPREF "KERN_DeviceNetLinkSocketIoControl: no genl_info\n"); return 0; }
  na = info->attrs[DRV_DEVICE_CMD]; if(!na) { printk(LOGPREF "KERN_DeviceNetLinkSocketIoControl: no genl_info attributes\n"); return 0; }
  mydata = (char *)nla_data(na); if(!mydata) { printk(LOGPREF "KERN_DeviceNetLinkSocketIoControl: no data\n"); return 0; }
  memset(KERN_DeviceMessage, 0, KERN_DeviceMessage_len); ioctl_num = *(unsigned int *)mydata; ioctl_size = *(unsigned int *)(mydata + sizeof(ioctl_num));

  if(ioctl_size > KERN_DeviceMessage_len) ioctl_size = KERN_DeviceMessage_len;
  memcpy(KERN_DeviceMessage, mydata + sizeof(ioctl_num), ioctl_size);
  Ami_Filter_DRV_IoControl_Function(ioctl_num, KERN_DeviceMessage, &ioctl_size);
  if(!(skb = genlmsg_new(nla_total_size(ioctl_size), GFP_ATOMIC))) { printk(LOGPREF "KERN_DeviceNetLinkSocketIoControl: genlmsg_new FAILED\n"); return 0; }
  if(!(msg_head = genlmsg_put(skb,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
    info->snd_portid,
#else
    info->snd_pid,
#endif
    info->snd_seq, &DRV_DEVICE_gnl_family, 0, DRV_DEVICE_CMD))) {
      nlmsg_free(skb); printk(LOGPREF "KERN_DeviceNetLinkSocketIoControl: genlmsg_put FAILED\n"); return 0;
  }
  if(nla_put(skb, DRV_DEVICE_CMD, ioctl_size, KERN_DeviceMessage)) { nlmsg_free(skb); printk(LOGPREF "KERN_DeviceNetLinkSocketIoControl: nla_put FAILED\n"); return 0; }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
  genlmsg_end(skb, msg_head); if(skb->len < 0)
#else
  if(genlmsg_end(skb, msg_head) < 0)
#endif
  { nlmsg_free(skb); printk(LOGPREF "KERN_DeviceNetLinkSocketIoControl: genlmsg_end FAILED\n"); return 0; }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
  if(genlmsg_unicast(genl_info_net(info), skb, info->snd_portid))
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
  if(genlmsg_unicast(genl_info_net(info), skb, info->snd_pid))
#else
  if(genlmsg_unicast(skb, info->snd_pid))
#endif
  { printk(LOGPREF "KERN_DeviceNetLinkSocketIoControl: genlmsg_unicast FAILED\n"); return 0; }
  return 0;
}
int __calling_convention KERN_RegisterDevice(int buf_len) {
  if(!(KERN_DeviceMessage = (char *)KERN_kmalloc(buf_len))) { printk(LOGPREF "KERN_DRV_DEVICERegisterDevice: can't allocate message buffer\n"); return 1; }
  KERN_DeviceMessage_len = (unsigned int)buf_len;
  strncpy(DRV_DEVICE_gnl_family.name, LIN_DEVNAME, sizeof(DRV_DEVICE_gnl_family.name));
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)) || (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 5))
  if(genl_register_family(&DRV_DEVICE_gnl_family)) { printk(LOGPREF "KERN_DRV_DEVICERegisterDevice: genl_register_family FAILED\n"); return 1; }
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0)) || ( (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 1)) && (RHEL_RELEASE_CODE <= RHEL_RELEASE_VERSION(7, 4)) )
  if(genl_register_family_with_ops(&DRV_DEVICE_gnl_family, DRV_DEVICE_gnl_ops)) { printk(LOGPREF "KERN_DRV_DEVICERegisterDevice: genl_register_family_with_ops FAILED\n"); return 1; }
#else
  if(genl_register_family(&DRV_DEVICE_gnl_family)) { printk(LOGPREF "KERN_DRV_DEVICERegisterDevice: genl_register_family FAILED\n"); return 1; }
  if(genl_register_ops(&DRV_DEVICE_gnl_family, DRV_DEVICE_gnl_ops)) { printk(LOGPREF "KERN_DRV_DEVICERegisterDevice: genl_register_ops FAILED\n"); return 1; }
#endif
  return 0;
}
void __calling_convention KERN_DeregisterDevice(void) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0)) || (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 1))
#else
  if(genl_unregister_ops(&DRV_DEVICE_gnl_family, DRV_DEVICE_gnl_ops)) { printk(LOGPREF "KERN_DRV_DEVICEDeregisterDevice: genl_unregister_ops FAILED\n"); return; }
#endif
  if(genl_unregister_family(&DRV_DEVICE_gnl_family)) { printk(LOGPREF "KERN_DRV_DEVICEDeregisterDevice: genl_unregister_family FAILED\n"); return; }
  if(KERN_DeviceMessage) KERN_kfree(KERN_DeviceMessage);
  KERN_DeviceMessage = 0; KERN_DeviceMessage_len = 0;
}
///// <- Device ***********************************************

///// -> Event ************************************************
enum { DRV_EVENT_CMD_UNSPEC, DRV_EVENT_CMD, DRV_EVENT_CMD_MAX };
static struct nla_policy DRV_EVENT_genl_policy[DRV_EVENT_CMD_MAX] = { {.type = NLA_BINARY
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)
,.validation_type = NLA_VALIDATE_NONE
#endif
}, {.type = NLA_BINARY
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)
,.validation_type = NLA_VALIDATE_NONE
#endif
}, };
static int KERN_EventNetLinkSocketIoControl(struct sk_buff *skb, struct genl_info *info) { (void)skb; (void)info; return 0; }
static struct genl_ops DRV_EVENT_gnl_ops[] = { { .cmd = DRV_EVENT_CMD, .flags = 0,
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 2, 0) && (RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(8, 0))
.policy = DRV_EVENT_genl_policy,
#endif
.doit = KERN_EventNetLinkSocketIoControl, .dumpit = NULL, }, };
static struct genl_family DRV_EVENT_gnl_family = { .hdrsize = 0,
.name = "", .version = VERSION_NR, .maxattr = DRV_EVENT_CMD_MAX - 1,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 12))
.module = THIS_MODULE,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 2, 0)
.policy = DRV_EVENT_genl_policy,
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)) || (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 2))
.ops = DRV_EVENT_gnl_ops, .n_ops = ARRAY_SIZE(DRV_EVENT_gnl_ops),
#else
.id = GENL_ID_GENERATE,
#endif
};
static int DRV_EVENT_seq = 0;
int __calling_convention KERN_SendEvent(int pid) {
  int                 event_code = 0;
  struct sk_buff     *skb;
  void               *msg_head;
  struct task_struct *ts; int tsfound = 0; if(pid) { for_each_process(ts) { if(ts->pid == pid) { tsfound++; break; } } }
  if(!tsfound) { printk(LOGPREF "KERN_DRV_EVENTSendEvent: listener process with PID %i not exists\n", pid); return 1; }
  if(!(skb = genlmsg_new(nla_total_size(sizeof(int)), GFP_ATOMIC))) { printk(LOGPREF "KERN_DRV_EVENTSendEvent: genlmsg_new FAILED\n"); return 1; }
  if(!(msg_head = genlmsg_put(skb, pid, ++DRV_EVENT_seq, &DRV_EVENT_gnl_family, 0, DRV_EVENT_CMD))) {
    nlmsg_free(skb); printk(LOGPREF "KERN_DRV_EVENTSendEvent: genlmsg_put FAILED\n"); return 1;
  }
  if(nla_put(skb, DRV_EVENT_CMD, sizeof(int), &event_code)) { nlmsg_free(skb); printk(LOGPREF "KERN_DRV_EVENTSendEvent: nla_put FAILED\n"); return 1; }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
  genlmsg_end(skb, msg_head); if(skb->len < 0)
#else
  if(genlmsg_end(skb, msg_head) < 0)
#endif
  { nlmsg_free(skb); printk(LOGPREF "KERN_DRV_EVENTSendEvent: genlmsg_end FAILED\n"); return 1; }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
  if(genlmsg_unicast(&init_net, skb, pid))
#else
  if(genlmsg_unicast(skb, pid))
#endif
  { printk(LOGPREF "KERN_DRV_EVENTSendEvent: genlmsg_unicast FAILED\n"); return 1; }
  return 0;
}
int __calling_convention KERN_ClearEvent(int pid) { (void)pid; return 0; /* No needed in Linux */ }
int __calling_convention KERN_RegisterEvent(void) {
  DRV_EVENT_seq = 0;
  strncpy(DRV_EVENT_gnl_family.name, LIN_EVENTNAME, sizeof(DRV_EVENT_gnl_family.name));
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)) || (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 5))
  if(genl_register_family(&DRV_EVENT_gnl_family)) { printk(LOGPREF "KERN_DRV_EVENTRegisterNetLinkSocket: genl_register_family FAILED\n"); return 1; }
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0)) || ( (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 1)) && (RHEL_RELEASE_CODE <= RHEL_RELEASE_VERSION(7, 4)) )
  if(genl_register_family_with_ops(&DRV_EVENT_gnl_family, DRV_EVENT_gnl_ops)) { printk(LOGPREF "KERN_DRV_EVENTRegisterNetLinkSocket: genl_register_family_with_ops FAILED\n"); return 1; }
#else
  if(genl_register_family(&DRV_EVENT_gnl_family)) { printk(LOGPREF "KERN_DRV_EVENTRegisterNetLinkSocket: genl_register_family FAILED\n"); return 1; }
  if(genl_register_ops(&DRV_EVENT_gnl_family, DRV_EVENT_gnl_ops)) { printk(LOGPREF "KERN_DRV_EVENTRegisterNetLinkSocket: genl_register_ops FAILED\n"); return 1; }
#endif
  return 0;
}
void __calling_convention KERN_DeregisterEvent(void) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0)) || (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 1))
#else
  if(genl_unregister_ops(&DRV_EVENT_gnl_family, DRV_EVENT_gnl_ops)) { printk(LOGPREF "KERN_DRV_EVENTDeregisterEvent: genl_unregister_ops FAILED\n"); return; }
#endif
  if(genl_unregister_family(&DRV_EVENT_gnl_family)) { printk(LOGPREF "KERN_DRV_EVENTDeregisterEvent: genl_unregister_family FAILED\n"); return; }
}
///// <- Event ************************************************

///// -> Device notifier **************************************
static int KERN_dev_notifier(struct notifier_block *self, unsigned long status, void *ptr) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
  struct net_device *dev = netdev_notifier_info_to_dev(ptr);
#else
  struct net_device *dev = (struct net_device *)ptr;
#endif

  (void)self;

  if(dev->type != ARPHRD_ETHER) return 0;

  switch(status) {
  case NETDEV_CHANGE: break;
  case NETDEV_UP:         Ami_DEVN_ChangeUPAdapter((KERN_netdevice)dev, dev->name); break;
  case NETDEV_DOWN:       Ami_DEVN_ChangeDOWNAdapter((KERN_netdevice)dev, dev->name); break;
  case NETDEV_CHANGEMTU:  Ami_DEVN_ChangeMtuAdapter((KERN_netdevice)dev, dev->name, dev->mtu); break;
  case NETDEV_CHANGEADDR: Ami_DEVN_ChangeAddrAdapter((KERN_netdevice)dev, dev->name, dev->addr_len, (char *)dev->perm_addr); break;
  case NETDEV_CHANGENAME: Ami_DEVN_ChangeNameAdapter((KERN_netdevice)dev, dev->name); break;
  case NETDEV_REGISTER:   Ami_DEVN_BindAdapter((KERN_netdevice)dev, dev->name, NIC_ETHERNET, dev->addr_len, (char *)dev->dev_addr, dev->mtu); break;
  case NETDEV_UNREGISTER: Ami_DEVN_UnBindAdapter((KERN_netdevice)dev, dev->name); break;
  default: break;
  }
  return 0;
}
static struct notifier_block KERN_devnb;
void __calling_convention KERN_register_device_notifier(void) {
  KERN_devnb.notifier_call = KERN_dev_notifier; KERN_devnb.next = NULL; KERN_devnb.priority = 0;
  register_netdevice_notifier(&KERN_devnb);
}
void __calling_convention KERN_unregister_device_notifier(void) { unregister_netdevice_notifier(&KERN_devnb); }
///// <- Device notifier **************************************

///// -> Power manager notifier *******************************
#include <linux/suspend.h>
static int KERN_pm_notifier(struct notifier_block *self, unsigned long status, void *ptr) {
  (void)self; (void)ptr;

  switch(status) {
  case PM_SUSPEND_PREPARE:
  case PM_HIBERNATION_PREPARE: Ami_PM_StandByPrepare(); break;
  case PM_POST_SUSPEND:
  case PM_POST_HIBERNATION:    Ami_PM_PostStandBy(); break;
  default: break;
  }
  return NOTIFY_DONE;
}
static struct notifier_block KERN_pmnb;
void __calling_convention KERN_register_pm_notifier(void) {
  KERN_pmnb.notifier_call = KERN_pm_notifier; KERN_pmnb.next = NULL; KERN_pmnb.priority = 0;
  register_pm_notifier(&KERN_pmnb);
}
void __calling_convention KERN_unregister_pm_notifier(void) { unregister_pm_notifier(&KERN_pmnb); }
///// <- Power manager notifier *******************************

///// -> File system ******************************************
#include <linux/fs.h>
const char * __calling_convention KERN_get_cfgfilename(unsigned int KERN_FILE_num)
{
  switch(KERN_FILE_num)
  {
  case KERN_FILE_RULES:      return LIN_RULESFNAME;
  case KERN_FILE_FSGUARDCFG: return LIN_FSGUARDCFGFNAME;
  case KERN_FILE_FSGUARDLOG: return LIN_FSGUARDLOGFNAME;
  }
  return NULL;
}
int __calling_convention KERN_fopen(const char *fname, KERN_file *file, int rewrite, int readonly) {
  struct file *FileHandle = filp_open(fname, rewrite ? (O_CREAT | O_RDWR | O_TRUNC) : (readonly ? O_RDONLY : O_RDWR), 0); *file = NULL;
  if(IS_ERR(FileHandle)) return -1;
  if(!FileHandle->f_op || !FileHandle->f_op->read || !FileHandle->f_op->write || !FileHandle->f_op->llseek) { filp_close(FileHandle, NULL); return -1; }
  *file = (KERN_file)FileHandle;
  return 0;
}
unsigned int __calling_convention KERN_fgetlength(KERN_file file) {
  if(!file) return -1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
  return (unsigned int)((struct file *)file)->f_path.dentry->d_inode->i_size;
#else
  return (unsigned int)((struct file *)file)->f_dentry->d_inode->i_size;
#endif
}
int __calling_convention KERN_fread(KERN_file file, unsigned int offset, void *buffer, unsigned int buflen) {
  size_t read = 0; if(!file) return -1;
  ((struct file *)file)->f_op->llseek((struct file *)file, (loff_t)offset, SEEK_SET);
  read = ((struct file *)file)->f_op->read((struct file *)file, buffer, buflen, &((struct file *)file)->f_pos);
  if(!read) return -1;
  return 0;
}
int __calling_convention KERN_fwrite(KERN_file file, unsigned int offset, void *buffer, unsigned int buflen) {
  size_t written = 0; if(!file) return -1;
  ((struct file *)file)->f_op->llseek((struct file *)file, (loff_t)offset, SEEK_SET);
  written = ((struct file *)file)->f_op->write((struct file *)file, buffer, buflen, &((struct file *)file)->f_pos);
  if(!written) return -1;
  return 0;
}
int __calling_convention KERN_fclose(KERN_file file) {
  if(!file) return -1;
  filp_close((struct file *)file, NULL);
  return 0;
}
int __calling_convention KERN_funlink(const char *fname)
{
  (void)fname;

  return 0; // !
}
int __calling_convention KERN_fnamelen(const char *fname)
{
  int fnamelen = 0;
  while(fname[fnamelen]) fnamelen++;
  return fnamelen;
}
int __calling_convention KERN_fnamecmp(const char *dst, unsigned int dstlen, const char *src, unsigned int srclen)
{
  if(srclen > dstlen) return 1;
  if(srclen < dstlen) return -1;
  return memcmp(src, dst, srclen);
}
///// <- File system ******************************************

///// -> File system filter ******************************************
void __calling_convention KERN_queue_fsguardcfgsave(void)
{
}
int __calling_convention KERN_fs_RegisterDevice(int buf_len)
{
  (void)buf_len;
  return 0;
}
///// <- File system filter ******************************************

module_init(KERN_init_module);
module_exit(KERN_cleanup_module);
MODULE_LICENSE     ("GPL"); //("Proprietary");
MODULE_AUTHOR      ("Amicon");
MODULE_DESCRIPTION ("FPSU-IP/Client VPN Filter driver");


