#ifndef _AMIKERNEL_H
#define _AMIKERNEL_H

// See userdef.h
#ifdef _MACOSX_
#define __calling_convention
#endif // _MACOSX_
#ifdef _LINUX_
#ifdef BUILD_AS_X64
#define __calling_convention __attribute__((regparm(0)))
#endif // BUILD_AS_X64
#ifdef BUILD_AS_X86
#define __calling_convention __attribute__((cdecl,regparm(0)))
#endif // BUILD_AS_X86
#ifdef BUILD_AS_ARM
#define __calling_convention
#endif // BUILD_AS_ARM
#endif // _LINUX_

#ifdef _WINDOWS_
#define inline
#endif

///// -> Strings **********************************************
#ifdef _WINDOWS_
void __calling_convention                                         KERN_printk(const char *fmt, ...);
void __calling_convention                                         KERN_snprintf(char *str, int str_len, const char *fmt, ...);
#else
void __calling_convention __attribute__ ((format (printf, 1, 2))) KERN_printk(const char *fmt, ...);
void __calling_convention __attribute__ ((format (printf, 3, 4))) KERN_snprintf(char *str, int str_len, const char *fmt, ...);
#endif
void __calling_convention KERN_strncpy(char *dst, const char *src, int dst_len);
void __calling_convention KERN_strncat(char *dst, const char *src, int dst_len);
int  __calling_convention KERN_strcmp (const char *dst, const char *src);
int  __calling_convention KERN_strlen (const char *str);
void __calling_convention KERN_memset (void *mem, char val, int size);
void __calling_convention KERN_memcpy (void *dst, const void *src, int size);
void __calling_convention KERN_memmove(void *dst, const void *src, int size);
///// <- Strings **********************************************

///// -> Types definitions ************************************
typedef void * KERN_netbuf;
typedef void * KERN_netdevice;
typedef void * KERN_netdeviceflt;
typedef void * KERN_spin_lock;
typedef void * KERN_file;
///// <- Types definitions ************************************

///// -> Module initialization ********************************
int  __calling_convention Ami_init_module   (void);
void __calling_convention Ami_cleanup_module(void);

int  __calling_convention Ami_FS_init_module   (void);
void __calling_convention Ami_FS_cleanup_module(void);
///// <- Module initialization ********************************

///// -> Network hooks ****************************************
unsigned int __calling_convention Ami_net_inputoutputEx         (int ForSend, KERN_netbuf netbuf, KERN_netdevice netdev, int CheckQueue);
unsigned int __calling_convention Ami_net_inputoutput           (int ForSend, KERN_netbuf netbuf, KERN_netdevice netdev);
void         __calling_convention Ami_net_inputoutput_CheckQueue(void);
void         __calling_convention Ami_net_inputoutput_complete  (int ForSend, KERN_netbuf netbuf, KERN_netdevice netdev);
void         __calling_convention KERN_net_register_hooks       (void);
void         __calling_convention KERN_net_unregister_hooks     (void);
///// <- Network hooks ****************************************

///// -> Memory ***********************************************
void * __calling_convention KERN_kmalloc(int size);
void   __calling_convention KERN_kfree  (void *pmem);
///// <- Memory ***********************************************

///// -> Network device ***************************************
int    __calling_convention KERN_netdev_getaddinfolen(void);
void * __calling_convention Ami_NETDEV_getaddinfo    (KERN_netdevice dev);
///// <- Network device ***************************************

///// -> Network buffer ***************************************
KERN_netbuf __calling_convention KERN_netbuf_alloc     (KERN_netdevice netdev, void *buffer, int buffer_len, int ForSend);
void        __calling_convention KERN_netbuf_free      (KERN_netbuf netbuf);
int         __calling_convention KERN_netbuf_len       (KERN_netbuf netbuf);
int         __calling_convention KERN_netbuf_getdata   (KERN_netbuf netbuf, char *buffer, int buffer_len);
#ifdef _WINDOWS_
#define KERN_netbuf_queue_send_MAXCOUNT 1 // Only one (other packets not visible...) for send!!!
#define KERN_netbuf_queue_recv_MAXCOUNT 1 // Only one (other packets not visible...) for recv!!!
#else
#define KERN_netbuf_queue_send_MAXCOUNT 1 // Only one, there are no lists
#define KERN_netbuf_queue_recv_MAXCOUNT 1 // Only one, there are no lists
#endif
int         __calling_convention KERN_netbuf_queue_send(KERN_netbuf *netbuf, int netbufcount, KERN_netdevice netdev);
int         __calling_convention KERN_netbuf_queue_recv(KERN_netbuf *netbuf, int netbufcount, KERN_netdevice netdev);
///// <- Network buffer ***************************************

///// -> Spin locks *******************************************
void __calling_convention KERN_spin_lock_init  (KERN_spin_lock *pLock);
void __calling_convention KERN_spin_lock_lock  (KERN_spin_lock *pLock);
void __calling_convention KERN_spin_lock_unlock(KERN_spin_lock *pLock);
void __calling_convention KERN_spin_lock_free  (KERN_spin_lock *pLock);
///// <- Spin locks *******************************************

///// -> Timer ************************************************
void __calling_convention KERN_InitTimer          (unsigned long timerval);
void __calling_convention KERN_DestroyTimer       (void);
void __calling_convention KERN_Sleep              (int ms);
void __calling_convention Ami_DEV_NetTimerCallback(void);
void __calling_convention KERN_DrvGetCurFtime     (int *sec, int *min, int *hour, int *mday, int *mon, int *year, unsigned long long *microsec);
///// <- Timer ************************************************

///// -> Device ***********************************************
int  __calling_convention KERN_RegisterDevice              (int buf_len);
void __calling_convention KERN_DeregisterDevice            (void);
int  __calling_convention Ami_Filter_DRV_IoControl_Function(unsigned int ioctl_num, char *Messag, int *Messag_len);
///// <- Device ***********************************************

///// -> Event ************************************************
int  __calling_convention KERN_SendEvent      (int pid);
int  __calling_convention KERN_ClearEvent     (int pid);
int  __calling_convention KERN_RegisterEvent  (void);
void __calling_convention KERN_DeregisterEvent(void);
///// <- Event ************************************************

///// -> Device notifier **************************************
void              __calling_convention KERN_register_device_notifier  (void);
void              __calling_convention KERN_unregister_device_notifier(void);
KERN_netdeviceflt __calling_convention Ami_DEVN_BindAdapter           (KERN_netdevice netdev, const char *devname, int type, int addr_len, const char *addr, int mtu);
#ifdef USE_WAN_ADAPTERS
typedef int (*KERN_CHECK_NETBUF_FOR_WANNETDEV_HANDLER)(KERN_netdeviceflt netdevflt, KERN_netdeviceflt wannetdevflt, KERN_netbuf netbuf, int ForSend);
KERN_netdeviceflt __calling_convention Ami_DEVN_BindWanAdapter        (KERN_netdeviceflt netdevflt, const char *devname, int type, int addr_len, const char *addr, int mtu,
                                                                       KERN_CHECK_NETBUF_FOR_WANNETDEV_HANDLER CheckWanFilter);
#endif
void              __calling_convention Ami_DEVN_UnBindAdapter         (KERN_netdevice netdev, const char *devname);
#ifdef USE_WAN_ADAPTERS
void              __calling_convention Ami_DEVN_UnBindWanAdapter      (KERN_netdeviceflt netdevflt, KERN_netdeviceflt wannetdevflt);
#endif
void              __calling_convention Ami_DEVN_ChangeNameAdapter     (KERN_netdevice netdev, const char *devname);
void              __calling_convention Ami_DEVN_ChangeAddrAdapter     (KERN_netdevice netdev, const char *devname, int addr_len, const char *addr);
void              __calling_convention Ami_DEVN_ChangeMtuAdapter      (KERN_netdevice netdev, const char *devname, int mtu);
void              __calling_convention Ami_DEVN_ChangeDOWNAdapter     (KERN_netdevice netdev, const char *devname);
void              __calling_convention Ami_DEVN_ChangeUPAdapter       (KERN_netdevice netdev, const char *devname);
///// <- Device notifier **************************************

///// -> Power manager notifier *******************************
void __calling_convention KERN_register_pm_notifier  (void);
void __calling_convention KERN_unregister_pm_notifier(void);
void __calling_convention Ami_PM_StandByPrepare      (void);
void __calling_convention Ami_PM_PostStandBy         (void);
///// <- Power manager notifier *******************************

///// -> File system ******************************************
enum
{
  KERN_FILE_RULES = 0,
  KERN_FILE_FSGUARDCFG,
  KERN_FILE_FSGUARDLOG,
  KERN_FILE_MAX
};
const char * __calling_convention KERN_get_cfgfilename(unsigned int KERN_FILE_num);
int          __calling_convention KERN_fopen      (const char *fname, KERN_file *file, int rewrite, int readonly);
unsigned int __calling_convention KERN_fgetlength (KERN_file file);
int          __calling_convention KERN_fread      (KERN_file file,  unsigned int offset, void *buffer, unsigned int buflen);
int          __calling_convention KERN_fwrite     (KERN_file file, unsigned int offset, void *buffer, unsigned int buflen);
int          __calling_convention KERN_fclose     (KERN_file file);
int          __calling_convention KERN_funlink    (const char *fname);
int          __calling_convention KERN_fnamelen   (const char *fname);
int          __calling_convention KERN_fnamecmp   (const char *dst, unsigned int dstlen, const char *src, unsigned int srclen);
///// <- File system ******************************************

///// -> File system filter ******************************************
void         __calling_convention KERN_queue_fsguardcfgsave(void);
int          __calling_convention KERN_fs_RegisterDevice(int buf_len);
///// <- File system filter ******************************************

#endif // _AMIKERNEL_H
